/*
 * Created on May 27, 2012
 */

package riven.virtualthreads.test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import craterstudio.func.Filter;

public class StaticGameItemGrid {
   public final int               x, y, w, h, cellSize, cols, rows;
   private final List<GameItem>[] grid;
   private final List<GameItem>   outer;
   
   public StaticGameItemGrid(int w, int h, int cellSize) {
      this(0, 0, w, h, cellSize);
   }
   
   public StaticGameItemGrid(int x, int y, int w, int h, int cellSize) {
      this.x = x;
      this.y = y;
      this.w = w;
      this.h = h;
      
      this.cellSize = cellSize;
      this.cols = w / cellSize + 1;
      this.rows = h / cellSize + 1;
      
      this.grid = new ArrayList[cols * rows];
      for (int i = 0; i < grid.length; i++) {
         this.grid[i] = new ArrayList<GameItem>();
      }
      this.outer = new ArrayList<GameItem>();
   }
   
   private List<GameItem> lookup(GameItem item) {
      int xi = (int) ((item.position.x - this.x) / this.cellSize);
      int yi = (int) ((item.position.y - this.y) / this.cellSize);
      
      if (xi < 0 || yi < 0 || xi >= this.cols || yi >= this.rows) {
         return this.outer;
      }
      return this.grid[yi * this.cols + xi];
   }
   
   public void put(GameItem item) {
      this.lookup(item).add(item);
   }
   
   public boolean remove(GameItem item) {
      return this.lookup(item).remove(item);
   }
   
   //
   
   public GameItem scan(Vec2 origin, float range, Filter<GameItem> filter) {
      int irangeMax = (int) Math.ceil(range / this.cellSize);
      
      ScanData scanData = new ScanData();
      
      for (int irange = 0; irange <= irangeMax; irange++) {
         
         this.scanImpl(origin, irange, filter, scanData);
         GameItem min = scanData.first(origin);
         if (min == null) {
            continue;
         }
         if (new Vec2(min.position).sub(origin).len2() > range * range) {
            return null;
         }
         return min;
      }
      
      return null;
   }
   
   private void scanImpl(final Vec2 origin, int range, Filter<GameItem> filter, ScanData scanData) {
      int xi = (int) ((origin.x - this.x) / this.cellSize);
      int yi = (int) ((origin.y - this.y) / this.cellSize);
      
      int x1 = Math.max(0, Math.min(xi - range, this.cols - 1));
      int x2 = Math.max(0, Math.min(xi + range, this.cols - 1));
      int y1 = Math.max(0, Math.min(yi - range, this.rows - 1));
      int y2 = Math.max(0, Math.min(yi + range, this.rows - 1));
      
      scanData.matches.clear();
      
      for (int y = y1; y <= y2; y++) {
         for (int x = x1; x <= x2; x++) {
            List<GameItem> cell = this.grid[y * this.cols + x];
            
            if (filter == null) {
               scanData.matches.addAll(cell);
            }
            else {
               for (GameItem item : cell) {
                  if (filter.accept(item)) {
                     scanData.matches.add(item);
                  }
               }
            }
         }
      }
   }
   
   static class ScanData {
      List<GameItem> matches = new ArrayList<GameItem>();
      GameItem[]     array   = new GameItem[16];
      
      void sort(final Vec2 origin) {
         while (matches.size() > this.array.length) {
            this.array = new GameItem[this.array.length * 2];
         }
         
         this.matches.toArray(this.array);
         
         Arrays.sort(this.array, 0, this.matches.size(), new Comparator<GameItem>() {
            Vec2 tmp = new Vec2();
            
            @Override
            public int compare(GameItem o1, GameItem o2) {
               float len1 = tmp.set(o1.position).sub(origin).len2();
               float len2 = tmp.set(o2.position).sub(origin).len2();
               int cmp = Integer.signum((int) (len1 - len2));
               return (cmp == 0) ? 1 : cmp;
            }
         });
         
         for (int i = 0; i < this.matches.size(); i++) {
            this.matches.set(i, this.array[i]);
         }
      }
      
      GameItem first(final Vec2 origin) {
         float minLen2 = Integer.MAX_VALUE;
         GameItem min = null;
         
         Vec2 tmp = new Vec2();
         
         for (GameItem item : this.matches) {
            float len2 = tmp.set(item.position).sub(origin).len2();
            if (len2 < minLen2) {
               minLen2 = len2;
               min = item;
            }
         }
         
         return min;
      }
   }
}
